﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using TextFieldParserStandard;

namespace InABox.Core
{
    public class VariableTypeLookupGenerator : LookupGenerator<DigitalFormVariable>
    {
        public VariableTypeLookupGenerator(DigitalFormVariable[] items) : base(items)
        {
            foreach (var fieldType in DFUtils.GetFieldTypes())
            {
                var caption = fieldType.GetCaption();
                if (string.IsNullOrWhiteSpace(caption))
                {
                    caption = CoreUtils.Neatify(fieldType.Name);
                }
                AddValue(fieldType.FullName, caption);
            }
        }
    }

    [Caption("Variables")]
    public class DigitalFormVariable : Entity, IRemotable, IPersistent, IStringAutoIncrement<DigitalFormVariable>, IOneToMany<DigitalForm>,
        ISequenceable, ILicense<DigitalFormsLicense>
    {
        private string _parameters = "";
        private DFLayoutFieldProperties _properties = new DFLayoutStringFieldProperties();

        private Type _variableType = typeof(DFLayoutStringField);

        [NullEditor]
        [Obsolete("Replaced with Form")]
        public QAFormLink QAForm { get; set; }

        [NullEditor]
        [EntityRelationship(DeleteAction.Cascade)]
        public DigitalFormLink Form { get; set; }

        [EditorSequence(1)]
        [CodeEditor(Visible = Visible.Default, Editable = Editable.Disabled)]
        public string Code { get; set; }

        [EditorSequence(2)]
        [TextBoxEditor(Visible = Visible.Default, Editable = Editable.Disabled)]
        public string Description { get; set; }

        [NullEditor]
        [Obsolete("Use VariableType Instead!")]
        public string Type { get; set; }

        [EditorSequence(3)]
        [ComboLookupEditor(typeof(VariableTypeLookupGenerator), Visible = Visible.Default)]
        public string VariableType
        {
            get => _variableType.FullName;
            set
            {
                _variableType = DFUtils.GetFieldType(value) ?? typeof(DFLayoutStringField);
                _properties = CreateProperties();
            }
        }
        
        [EditorSequence(7)]
        [MemoEditor(Visible = Visible.Hidden, Editable = Editable.Disabled)]
        public string Parameters
        {
            get => _parameters;
            set
            {
                _parameters = value;
                _properties = CreateProperties();
            }
        }

        [CheckBoxEditor(Visible = Visible.Default, Editable = Editable.Disabled)]
        [EditorSequence(8)]
        public bool Required { get; set; }

        [CheckBoxEditor(Visible = Visible.Default, Editable = Editable.Disabled)]
        [EditorSequence(9)]
        public bool Secure { get; set; }

        [CheckBoxEditor(Visible = Visible.Default, Editable = Editable.Disabled)]
        [EditorSequence(10)]
        public bool Retain { get; set; }

        [NullEditor]
        public long Sequence { get; set; }

        public Expression<Func<DigitalFormVariable, string>> AutoIncrementField()
        {
            return x => x.Code;
        }

        public Filter<DigitalFormVariable> AutoIncrementFilter()
        {
            return new Filter<DigitalFormVariable>(x => x.Form.ID).IsEqualTo(Form.ID);
        }

        public string AutoIncrementFormat()
        {
            return "V{0:D3}";
        }

        public object ParseValue(object value)
        {
            return _properties.ParseValue(value);
        }

        public string FormatValue(object value)
        {
            return _properties.FormatValue(value);
        }

        public static Type DataType(Type variableType)
        {
            var propertiesType = GetPropertiesType(variableType);
            var baseType = propertiesType?.GetSuperclassDefinition(typeof(DFLayoutFieldProperties<>));
            var dataType = baseType?.GenericTypeArguments[0] ?? typeof(string);
            return dataType;
        }

        public Type FieldType()
        {
            return _variableType;
        }

        protected override void Init()
        {
            base.Init();
            QAForm = new QAFormLink();
            Form = new DigitalFormLink();
        }

        public static Type? GetPropertiesType(Type fieldType)
        {
            var fieldSuperType = fieldType.GetSuperclassDefinition(typeof(DFLayoutField<>));
            if (fieldSuperType != null)
            {
                return fieldSuperType.GenericTypeArguments[0];
            }
            return null;
        }

        public DFLayoutFieldProperties CreateProperties()
        {
            var propertiesType = GetPropertiesType(_variableType) ?? typeof(DFLayoutStringFieldProperties);

            var properties = (Activator.CreateInstance(propertiesType) as DFLayoutFieldProperties)!;
            properties.LoadFromString(Parameters);
            properties.Code = Code;
            return properties;
        }

        public DFLayoutFieldProperties LoadProperties(DFLayoutField field)
        {
            DFLayoutFieldProperties? properties = null;
            if (field.GetType().IsSubclassOfRawGeneric(typeof(DFLayoutField<>)))
            {
                properties = field.GetType().GetProperty("Properties").GetValue(field) as DFLayoutFieldProperties;
            }
            if(properties != null)
            {
                properties.LoadFromString(Parameters);
            }
            return properties;
        }

        public void SaveProperties(Type variableType, DFLayoutFieldProperties properties)
        {
            _variableType = variableType;
            SaveProperties(properties);
        }

        public void SaveProperties(DFLayoutFieldProperties properties)
        {
            Code = properties.Code;
            Description = properties.Description;
            Required = properties.Required;
            Secure = properties.Secure;
            Retain = properties.Retain;
            Parameters = properties.SaveToString();
        }
    }
}