﻿using InABox.Core;
using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using Xceed.Wpf.Toolkit.PropertyGrid.Converters;

namespace InABox.DynamicGrid
{
    public delegate void FieldChangedEvent();

    public interface IDynamicFormFieldControl
    {
        public event FieldChangedEvent? FieldChangedEvent;

        public object? GetValue();
        /// <summary>
        /// Sets the value in this control.
        /// </summary>
        /// <param name="value">The value to set. This will be the return value from a call to <see cref="DFLayoutFieldProperties.ParseValue(object)"/>.</param>
        public void SetValue(object? value);

        public object? GetEntityValue();
        public void SetEntityValue(object? value);

        /// <summary>
        /// Check that the data is valid - if it is not, output a message for the user.
        /// This function gets called when the user completes a form, or edits an already completed form.
        /// </summary>
        /// <param name="message">The message to the user.</param>
        /// <returns><see langword="true"/> if the data is valid.</returns>
        public bool Validate([NotNullWhen(false)] out string? message);
    }

    public abstract class DynamicFormFieldControl<TField, TProperties, TValue> : DynamicFormControl<TField>, IDynamicFormFieldControl
        where TField : DFLayoutField<TProperties>
        where TProperties : DFLayoutFieldProperties<TValue>, new()
    {
        public event FieldChangedEvent? FieldChangedEvent;

        public TField Field { get => Control; set => Control = value; }

        protected void ChangeField() => FieldChangedEvent?.Invoke();

        /// <summary>
        /// Checks whether the user has supplied a field - for use with <see cref="Validate(out string?)"/>.
        /// </summary>
        /// <returns><see langword="true"/> if the user has not supplied a value.</returns>
        protected abstract bool IsEmpty();

        public virtual bool Validate([NotNullWhen(false)] out string? message)
        {
            if(Field.Properties.Required && IsEmpty())
            {
                message = $"Field [{Field.Name}] is required!";
                return false;
            }
            message = null;
            return true;
        }

        public abstract TValue GetValue();
        public abstract void SetValue(TValue? value);

        public virtual object? GetEntityValue() => GetValue();
        public virtual void SetEntityValue(object? value) => SetValue(CoreUtils.ChangeType<TValue>(value));

        object? IDynamicFormFieldControl.GetValue() => GetValue();
        void IDynamicFormFieldControl.SetValue(object? value) => SetValue(value != null ? (TValue)value : default);
    }
}
